#!/usr/bin/env nix-shell
#!nix-shell -i python -p "python3.withPackages(ps : with ps; [ pyyaml paho-mqtt ])"

import time
import random
from MMQTT import MMQTT


class Device:

  def __init__(self, name: str, mqtt: MMQTT):
    self.mqtt = mqtt
    self.name = name

    # publish online state…

  def topic(self, subtopic: str):
    '''subtopic mustn't start with /'''

    return f'device/{self.name}/{subtopic}'

  def publish(self, subtopic: str, payload=None, qos=0, retain=False, prettyPrintYAML=False):
    '''subtopic mustn't start with /'''

    self.mqtt.publish(topic=self.topic(subtopic), payload=payload, qos=qos, retain=retain, prettyPrintYAML=prettyPrintYAML)


class Action:
  def execute(self, dev: Device) -> None: pass


class SleepAction(Action):
  def __init__(self, seconds: float = 0.1337):
    self.time = seconds
  def execute(self, dev: Device) -> None:
    time.sleep(self.time)


class PublishAction(Action):

  def __init__(self, subtopic: str, payload=None, qos=0, retain=False, prettyPrintYAML=False):
    self.qos = qos
    self.retain = retain
    self.payload = payload
    self.subtopic = subtopic
    self.prettyPrintYAML = prettyPrintYAML

  def execute(self, dev: Device) -> None:
    dev.publish(subtopic=self.subtopic, payload=self.payload, qos=self.qos, retain=self.retain, prettyPrintYAML=self.prettyPrintYAML)


class OnlineStatusAction:

  def __init__(self, online: bool):
    self.online = online

  def execute(self, dev: Device) -> None:
    dev.publish(
      subtopic = 'status',
      payload = 'online' if self.online else None,
      retain = self.online
    )


class ListenerAction:

  def __init__(self, ipv4: str = 'localhost', port: int = 4213):
    self.ipv4 = ipv4
    self.port = port

  def execute(self, dev: Device) -> None:

    listener = {
      'IPv4': self.ipv4,
      'port': self.port
    }

    dev.publish(
      subtopic = 'listener',
      payload = listener
    )


class MetricsAction:

  def __init__(self, nameSpace: str = 'CyanLight', reset: int = (1, 1337), uptime: int = (1, 1e6), lifetime: int = None, birth: int = (1e9, time.time()), **metrics):
    '''Setting a metric to None means it gets randomized every action
    
    The predefined metrics are mandatory, but may be overwritten through **kwargs
    The randomized metrics may be further specified by submitting a 2-tuple instead of None'''

    self.nameSpace = nameSpace

    self.metrics = {}
    self.metrics['birth'] = birth
    self.metrics['reset'] = reset
    self.metrics['uptime'] = uptime
    self.metrics['lifetime'] = lifetime

    for k, v in metrics.items():
      self.metrics[k] = v

  def execute(self, dev: Device) -> None:

    metrics = {}

    for k, v in self.metrics.items():
      if v is None:
        metrics[k] = random.randrange(1, 2**32)
      elif issubclass(type(v), tuple) and len(v) >= 2 and v[0] < v[1]:
        metrics[k] = random.randrange(int(v[0]), int(v[1]))
      else:
        metrics[k] = v

    dev.publish(
      subtopic = f'{self.nameSpace}/metrics',
      payload = metrics
    )



mqtt = MMQTT()

if __name__ == '__main__':
  # start main program

  on = OnlineStatusAction(True)
  off = OnlineStatusAction(False)
  p1 = SleepAction(1)
  p3 = SleepAction(3)
  p5 = SleepAction(5)
  # p10 = SleepAction(10)
  listener = ListenerAction()
  metrics = MetricsAction()

  dev = Device('ringclock', mqtt)

  while True:
    on.execute(dev)
    p1.execute(dev)
    listener.execute(dev)
    p1.execute(dev)
    for _ in range(10):
      metrics.execute(dev)
      p3.execute(dev)
    p3.execute(dev)
    off.execute(dev)
    p5.execute(dev)
